"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import glob
import json
import logging
from argparse import Namespace
from typing import Optional, Union, Any, List
from lxml import etree as ElementTree
import libvirt
from libvirtnbdbackup import output
from libvirtnbdbackup.virt import xml
from libvirtnbdbackup.output.exceptions import OutputException
from libvirtnbdbackup.common import defaultCheckpointName
from libvirtnbdbackup.exceptions import (
    NoCheckpointsFound,
    ReadCheckpointsError,
    CheckpointException,
    SaveCheckpointError,
    ForeignCeckpointError,
    RedefineCheckpointError,
    RemoveCheckpointError,
)

log = logging.getLogger()

redefineFlags = (
    libvirt.VIR_DOMAIN_CHECKPOINT_CREATE_REDEFINE
    | libvirt.VIR_DOMAIN_CHECKPOINT_CREATE_REDEFINE_VALIDATE
)


def exists(
    domObj: libvirt.virDomain, checkpointName: str
) -> libvirt.virDomainCheckpoint:
    """Check if an checkpoint exists"""
    return domObj.checkpointLookupByName(checkpointName)


def getXml(cptObj: libvirt.virDomainCheckpoint) -> str:
    """Get Checkpoint XML including size, if possible. Flag
    is not supported amongst all libvirt versions."""
    try:
        return cptObj.getXMLDesc(libvirt.VIR_DOMAIN_CHECKPOINT_XML_SIZE)
    except libvirt.libvirtError as e:
        log.warning("Failed to get checkpoint info with size information: [%s]", e)
        return cptObj.getXMLDesc()


def getSize(domObj: libvirt.virDomain, checkpointName: str) -> int:
    """Return current size of checkpoint for all disks"""
    size: int = 0
    cpt = exists(domObj, checkpointName)
    cptTree = xml.asTree(getXml(cpt))
    for s in cptTree.xpath("disks/disk/@size"):
        size += int(s)

    return size


def delete(domObj: libvirt.virDomain, cptObj: libvirt.virDomainCheckpoint) -> bool:
    """Delete checkpoint or checkpoint metadata in case validation
    fails (bitmap is missing, but checkpoint still existent)"""
    flags: int = 0
    checkpointName = cptObj.getName()
    if defaultCheckpointName not in checkpointName:
        log.debug(
            "Skipping checkpoint removal: [%s]: not from this application",
            checkpointName,
        )
        return True

    if not validate(domObj, checkpointName):
        log.warning(
            "Checkpoint inconsistency detected, removing metadata for checkpoint."
        )
        flags = libvirt.VIR_DOMAIN_CHECKPOINT_DELETE_METADATA_ONLY

    try:
        log.debug("Attempt to remove checkpoint: [%s]", checkpointName)
        cptObj.delete(flags)
        log.debug("Removed checkpoint: [%s]", checkpointName)
        return True
    except libvirt.libvirtError as errmsg:
        log.error("Error during checkpoint removal: [%s]", errmsg)
        return False


def backup(args: Namespace, domObj: libvirt.virDomain) -> bool:
    """save checkpoint config to persistent storage"""
    checkpointFile = os.path.join(args.checkpointdir, f"{args.cpt.name}.xml")
    log.info("Saving checkpoint config to: [%s]", checkpointFile)
    try:
        with output.openfile(checkpointFile, "wb") as f:
            c = exists(domObj, args.cpt.name)
            f.write(getXml(c).encode())
            return True
    except OutputException as errmsg:
        log.error(
            "Failed to save checkpoint config to file: [%s]: %s",
            checkpointFile,
            errmsg,
        )
        return False


def _hasForeign(domObj: libvirt.virDomain) -> Optional[str]:
    """Check if the virtual machine has an checkpoint which was not
    created by virtnbdbackup

    If an user or a third party utility creates an checkpoint,
    it is in line with the complete checkpoint chain, but
    virtnbdbackup does not save it. We can ensure consistency
    only if the complete chain of checkpoints is created by
    ourself. In case we detect an checkpoint that does not
    match our name, return so.
    """
    cpts = domObj.listAllCheckpoints()
    if not cpts:
        return None

    for cpt in cpts:
        checkpointName = cpt.getName()
        if defaultCheckpointName not in checkpointName:
            return checkpointName

    return None


def checkForeign(
    args: Namespace,
    domObj: libvirt.virDomain,
) -> bool:
    """Check and warn user if virtual machine has checkpoints
    not originating from this utility"""
    foreign = None
    if args.level in ("full", "inc", "diff"):
        foreign = _hasForeign(domObj)

    if not foreign:
        return True

    log.fatal("Foreign checkpoint found: [%s]", foreign)
    log.fatal("This checkpoint has not been created by this utility.")
    log.fatal(
        "To ensure backup chain consistency, "
        "remove existing checkpoints "
        "and start a new backup chain by creating a full backup."
    )

    raise ForeignCeckpointError


def removeAll(
    domObj: libvirt.virDomain,
    checkpointList: Union[List[Any], None],
    args: Namespace,
) -> bool:
    """Remove all existing checkpoints for a virtual machine,
    used during FULL backup to reset checkpoint chain, either
    by using the list from the checkpoint file (for transient
    checkpoint storage using --checkpointdir) or by using the
    checkpoints as listed by the libvirt API
    """
    log.info("Removing all existent checkpoints before full backup.")
    try:
        log.debug("Cleaning up persistent storage %s", args.checkpointdir)
        for checkpointFile in glob.glob(os.path.join(args.checkpointdir, "*.xml")):
            log.debug("Remove checkpoint file: %s", checkpointFile)
            os.remove(checkpointFile)
    except OSError as e:
        log.error("Failed to clean persistent storage %s: %s", args.checkpointdir, e)
        return False

    if checkpointList is None:
        cpts = domObj.listAllCheckpoints()
        if not cpts:
            return True
        for cpt in cpts:
            if delete(domObj, cpt) is False:
                return False
        return True

    for cp in checkpointList:
        cptObj = exists(domObj, cp)
        if cptObj:
            if delete(domObj, cptObj) is False:
                return False
    return True


def redefine(domObj: libvirt.virDomain, args: Namespace) -> bool:
    """Redefine checkpoints from persistent storage"""
    checkpointList = glob.glob(f"{args.checkpointdir}/*.xml")
    checkpointList.sort(key=os.path.getmtime)

    if checkpointList:
        log.info("Loaded checkpoint list from: [%s]", args.checkpointdir)

    for checkpointFile in checkpointList:
        try:
            log.debug("Loading checkpoint config from: [%s]", checkpointFile)
            with output.openfile(checkpointFile, "rb") as f:
                checkpointConfig = f.read()
                root = ElementTree.fromstring(checkpointConfig)
        except OutputException as e:
            log.error("Opening checkpoint file failed: [%s]: %s", checkpointFile, e)
            return False
        except ElementTree.ParseError as e:
            log.error(
                "Failed to load checkpoint config from [%s]: %s", checkpointFile, e
            )
            return False

        try:
            checkpointName = root.find("name").text
        except ElementTree.ParseError as e:
            log.error("Failed to find checkpoint name: [%s]", e)
            return False

        try:
            _ = exists(domObj, checkpointName)
            log.debug("Checkpoint [%s] found", checkpointName)
            continue
        except libvirt.libvirtError as e:
            # ignore VIR_ERR_NO_DOMAIN_CHECKPOINT, report other errors
            if e.get_error_code() != libvirt.VIR_ERR_NO_DOMAIN_CHECKPOINT:
                log.error("libvirt error: %s", e)
                return False

        log.info("Redefine missing checkpoint: [%s]", checkpointName)
        try:
            domObj.checkpointCreateXML(
                checkpointConfig.decode(),
                redefineFlags,
            )
        except libvirt.libvirtError as e:
            log.error("Redefining checkpoint failed: [%s]: %s", checkpointName, e)
            return False

    return True


def read(cFile: str) -> List[str]:
    """Open checkpoint file and read checkpoint
    information"""
    checkpoints: List[str] = []
    if not os.path.exists(cFile):
        return checkpoints

    try:
        with output.openfile(cFile, "rb") as fh:
            checkpoints = json.loads(fh.read().decode())
        return checkpoints
    except OutputException as e:
        raise ReadCheckpointsError(f"Failed to read checkpoint file: [{e}]") from e
    except json.decoder.JSONDecodeError as e:
        raise ReadCheckpointsError(f"Invalid checkpoint file: [{e}]") from e


def save(args: Namespace) -> None:
    """Append created checkpoint to checkpoint
    file"""
    try:
        checkpoints = read(args.cpt.file)
        checkpoints.append(args.cpt.name)
        with output.openfile(args.cpt.file, "wb") as cFw:
            cFw.write(json.dumps(checkpoints).encode())
    except CheckpointException as e:
        raise CheckpointException from e
    except OutputException as e:
        raise SaveCheckpointError from e


def validate(domObj: libvirt.virDomain, checkpointName: str) -> bool:
    """Validate that checkpoint and bitmap exist and are OK
    Currently the only way to correctly verify consistency is by redefine the checkpoint with
    VIR_DOMAIN_CHECKPOINT_CREATE_REDEFINE_VALIDATE option set, which will check the bitmap
    consistency, too."""
    try:
        c = domObj.checkpointLookupByName(checkpointName)
        checkpointXml = c.getXMLDesc(0)

        # Redefine the checkpoint using the provided XML description
        if not domObj.checkpointCreateXML(checkpointXml, redefineFlags):
            return False
        return True
    except libvirt.libvirtError as e:
        log.warning("Failed to validate checkpoint: [%s]", e)
        return False


def create(
    args: Namespace,
    domObj: libvirt.virDomain,
) -> None:
    """Checkpoint handling for different backup modes
    to be executed. Create, check and redefine checkpoints based
    on backup mode.

    Creates a new namespace in the argparse object,
    for easy pass around in further functions.
    """
    checkpointName: str = f"{defaultCheckpointName}.0"
    parentCheckpoint: str = ""
    cptFile: str = os.path.join(args.output, f"{args.domain}.cpt")
    checkpoints: List[str] = read(cptFile)

    if args.offline is False:
        if redefine(domObj, args) is False:
            raise RedefineCheckpointError("Failed to redefine checkpoints.")

    # save level to reuse it properly when filename is selected for the backup
    args.level_filename = args.level

    # remove checkpoints as loaded from the checkpoint json file
    if args.level == "full" and checkpoints:
        log.info("Loaded checkpoints from: [%s]", cptFile)
        if not removeAll(domObj, checkpoints, args):
            raise RemoveCheckpointError("Failed to remove checkpoint.")
        os.remove(cptFile)
        checkpoints = []
    # if no checkpoints are found in file, remove all existent ones matching the
    # name
    elif args.level == "full" and len(checkpoints) < 1:
        if not removeAll(domObj, None, args):
            raise RemoveCheckpointError("Failed to remove checkpoint.")
        checkpoints = []

    if checkpoints and args.level in ("inc", "diff"):
        nextCpt = len(checkpoints)
        checkpointName = f"{defaultCheckpointName}.{nextCpt}"
        parentCheckpoint = checkpoints[-1]
        log.info("Next checkpoint id: [%s].", nextCpt)
        log.info("Parent checkpoint name [%s].", parentCheckpoint)

        if args.offline is True:
            log.info("Offline backup, using latest checkpoint, saving only delta.")
            checkpointName = parentCheckpoint
        # Autostart is disabled, validation could not be performed
        elif not validate(domObj, parentCheckpoint):
            log.warning(
                "Checkpoint [%s] is invalid, switching backup to full.",
                parentCheckpoint,
            )
            args.level = "full"
            # reset back to defaults
            parentCheckpoint = ""

    if args.level in ("inc", "diff") and len(checkpoints) < 1:
        raise NoCheckpointsFound(
            "No existing checkpoints found, execute full backup first."
        )

    if args.level == "diff":
        log.info("Diff backup: saving delta since checkpoint: [%s].", parentCheckpoint)

    if args.level in ("full", "inc"):
        log.info("Using checkpoint name: [%s].", checkpointName)

    args.cpt = Namespace()
    args.cpt.name = checkpointName
    args.cpt.parent = parentCheckpoint
    args.cpt.file = cptFile

    log.debug("Checkpoint info: [%s].", vars(args.cpt))



#!/usr/bin/python3
"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import json
import logging
from argparse import Namespace
from typing import List, Union
import libvirt

from libvirtnbdbackup import output
from libvirtnbdbackup.virt.client import DomainDisk
from libvirtnbdbackup import common as lib
from libvirtnbdbackup.qemu import util as qemu
from libvirtnbdbackup.qemu.exceptions import ProcessError
from libvirtnbdbackup.ssh.exceptions import sshError
from libvirtnbdbackup.output.exceptions import OutputException
from libvirtnbdbackup.output.target.directory import Directory
from libvirtnbdbackup.output.target.zip import Zip
from libvirtnbdbackup.common import safeInfo
from libvirtnbdbackup.virt import guest


log = logging.getLogger()


def backupChecksum(fileStream, targetFile):
    """Save the calculated adler32 checksum, it can be verified
    by virtnbdbrestore's verify function.'"""
    checksum = fileStream.checksum()
    safeInfo("Checksum for file: [%s]:[%s]", targetFile, checksum)
    chksumfile = f"{targetFile}.chksum"
    safeInfo("Saving checksum to: [%s]", chksumfile)
    with output.openfile(chksumfile, "w") as cf:
        cf.write(f"{checksum}")


def backupConfig(args: Namespace, vmConfig: str) -> Union[str, None]:
    """Save domain XML config file"""
    configFile = os.path.join(args.output, f"vmconfig.{lib.getIdent(args)}.xml")
    log.info("Saving VM config to: [%s]", configFile)
    try:
        with output.openfile(configFile, "wb") as fh:
            fh.write(vmConfig.encode())
        return configFile
    except OutputException as e:
        log.error("Failed to save VM config: [%s]", e)
        return None


def backupDiskInfo(args: Namespace, disk: DomainDisk):
    """Save information about qcow image, used to reconstruct
    the qemu image with the same settings during restore"""
    try:
        info = qemu.util("").info(disk.path, args.sshClient)
    except (
        ProcessError,
        sshError,
    ) as errmsg:
        log.warning("Failed to read qcow image info: [%s]", errmsg)
        return

    configFile = f"{args.output}/{disk.target}.{lib.getIdent(args)}.qcow.json"
    try:
        with output.openfile(configFile, "wb") as fh:
            fh.write(info.out.encode())
        log.info("Saved qcow image config to: [%s]", configFile)
        if args.stdout is True:
            args.diskInfo.append(configFile)
    except OutputException as e:
        log.warning("Failed to save qcow image config: [%s]", e)


def backupBootConfig(args: Namespace) -> None:
    """Save domain uefi/nvram/kernel and loader if configured."""
    for setting, val in args.info.items():
        if args.level != "copy":
            tFile = os.path.join(
                args.output, f"{os.path.basename(val)}.{lib.getIdent(args)}"
            )
        else:
            tFile = os.path.join(args.output, f"{os.path.basename(val)}")
        log.info("Save additional boot config [%s] to: [%s]", setting, tFile)
        lib.copy(args, val, tFile)
        args.info[setting] = tFile


def backupAutoStart(args: Namespace) -> None:
    """Save information if virtual machine was marked
    for autostart during system boot"""
    log.info("Autostart setting configured for virtual machine.")
    autoStartFile = os.path.join(args.output, f"autostart.{lib.getIdent(args)}")
    try:
        with output.openfile(autoStartFile, "wb") as fh:
            fh.write(b"True")
    except OutputException as e:
        log.warning("Failed to save autostart information: [%s]", e)


def backupGuestInfo(args: Namespace) -> None:
    """Save OS related information"""
    osInfoFile = os.path.join(args.output, f"osinfo.{lib.getIdent(args)}")
    try:
        with output.openfile(osInfoFile, "w") as fh:
            fh.write(json.dumps(args.guestInfo, indent=4))
        log.info("Saved guest related osinfo to [%s]", osInfoFile)
    except OutputException as e:
        log.warning("Failed to save osinfo data: [%s]", e)


def backupBitlockerRecoveryKey(args: Namespace, domObj: libvirt.virDomain) -> None:
    """Save bitlocker recovery keys"""
    try:
        bde = guest.Exec(domObj, "manage-bde.exe", ["-status"])
        log.info("Bitlocker tools detected, attempting to backup recovery keys.")
        log.debug(bde)
    except libvirt.libvirtError:
        log.info("System does not appear to have bitlocker tools installed, skipping.")

    for i in range(0, args.guestInfo["fs.count"]):
        vol = args.guestInfo.get(f"fs.{i}.mountpoint", None)
        if not vol:
            continue
        if not ":" in vol:
            log.info("Skipping volume: [%s]", vol)
            continue

        vol = vol.replace("\\", "")
        log.info("Get bitlocker recovery key for volume: [%s]", vol)
        try:
            protectors = guest.Exec(
                domObj, "manage-bde.exe", ["-protectors", "-get", vol]
            )
        except RuntimeError:
            log.warning(
                "Unable to extract recovery key for volume [%s], see debug log for error details.",
                vol,
            )
            continue
        except (TimeoutError, libvirt.libvirtError) as e:
            log.warning("Unable to pull recovery keys for [%s]: %s", vol, e)
            continue

        keyFile = os.path.join(
            args.output,
            f"bitlocker.recovery.key.{vol.replace(':','')}.{lib.getIdent(args)}",
        )
        try:
            with output.openfile(keyFile, "w") as fh:
                fh.write(protectors)
            log.info("Saved Bitlocker recovery key to [%s]", keyFile)
        except OutputException as e:
            log.warning("Failed to save recovery key: [%s]", e)


def saveFiles(
    args: Namespace,
    vmConfig: str,
    disks: List[DomainDisk],
    fileStream: Union[Directory, Zip],
    logFile: str,
):
    """Save additional files such as virtual machine configuration
    and UEFI / kernel images"""
    configFile = backupConfig(args, vmConfig)

    backupBootConfig(args)
    for disk in disks:
        if disk.format.startswith("qcow"):
            backupDiskInfo(args, disk)
    if args.stdout is True:
        addFiles(args, configFile, fileStream, logFile)


def addFiles(args: Namespace, configFile: Union[str, None], zipStream, logFile: str):
    """Add backup log and other files to zip archive"""
    if configFile is not None:
        log.info("Adding vm config to zipfile")
        zipStream.zipStream.write(configFile, configFile)
    if args.level in ("full", "inc"):
        log.info("Adding checkpoint info to zipfile")
        zipStream.zipStream.write(args.cpt.file, args.cpt.file)
        for dirname, _, files in os.walk(args.checkpointdir):
            zipStream.zipStream.write(dirname)
            for filename in files:
                zipStream.zipStream.write(os.path.join(dirname, filename))

    for setting, val in args.info.items():
        log.info("Adding additional [%s] setting file [%s] to zipfile", setting, val)
        zipStream.zipStream.write(val, os.path.basename(val))

    for diskInfo in args.diskInfo:
        log.info("Adding QCOW image format file [%s] to zipfile", diskInfo)
        zipStream.zipStream.write(diskInfo, os.path.basename(diskInfo))

    log.info("Adding backup log [%s] to zipfile", logFile)
    zipStream.zipStream.write(logFile, logFile)



#!/usr/bin/env python3
"""Setup virtnbdbackup"""
from setuptools import setup, find_packages

import libvirtnbdbackup

with open("requirements.txt") as f:
    install_requires = f.read().splitlines()

setup(
    name="virtnbdbackup",
    version=libvirtnbdbackup.__version__,
    description="Backup utility for libvirt",
    url="https://github.com/abbbi/virtnbdbackup/",
    author="Michael Ablassmeier",
    author_email="abi@grinser.de",
    license="GPL",
    keywords="libnbd backup libvirt",
    packages=find_packages(exclude=("docs", "tests", "env")),
    include_package_data=True,
    scripts=["virtnbdbackup", "virtnbdrestore", "virtnbdmap", "virtnbd-nbdkit-plugin"],
    install_requires=install_requires,
    extras_require={
        "dev": [],
        "docs": [],
        "testing": [],
    },
    classifiers=[],
)



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys
import glob
import json
import logging
import logging.handlers
import signal
import shutil
import pprint
from time import time
from threading import current_thread
from argparse import Namespace
from typing import Optional, List, Any, Union, Dict
from tqdm import tqdm
import colorlog

from libvirtnbdbackup import ssh
from libvirtnbdbackup.ssh.exceptions import sshError
from libvirtnbdbackup import output
from libvirtnbdbackup.logcount import logCount

log = logging.getLogger("lib")

logFormat = (
    "%(asctime)s %(levelname)s %(name)s %(module)s - %(funcName)s"
    " [%(threadName)s]: %(message)s"
)
logFormatColored = (
    "%(green)s%(asctime)s%(reset)s%(blue)s %(log_color)s%(levelname)s%(reset)s "
    "%(name)s %(module)s - %(funcName)s"
    " [%(threadName)s]: %(log_color)s %(message)s"
)

logDateFormat = "[%Y-%m-%d %H:%M:%S]"
defaultCheckpointName = "virtnbdbackup"


def argparse(parser) -> Namespace:
    """Parse arguments"""
    return parser.parse_args()


def printVersion(version) -> None:
    """Print version and passed arguments"""
    log.info("Version: %s Arguments: %s", version, " ".join(sys.argv))


def humanize(num, suffix="B"):
    """Print size in human readable output"""
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def setThreadName(tn="main") -> None:
    """Set thread name reported by logging function"""
    current_thread().name = tn


def setLogLevel(verbose: bool) -> int:
    """Set loglevel"""
    level = logging.INFO
    if verbose is True:
        level = logging.DEBUG

    return level


def sshSession(
    args: Namespace, remoteHost: str, mode: ssh.Mode = ssh.Mode.DOWNLOAD
) -> Union[ssh.client, None]:
    """Use ssh to copy remote files"""
    try:
        return ssh.client(remoteHost, args.ssh_user, args.ssh_port, mode)
    except sshError as err:
        log.warning("Failed to setup SSH connection: [%s]", err)

    return None


def getLogFile(fileName: str) -> Optional[logging.FileHandler]:
    """Try setup log handler, if this fails, something is already
    wrong, but we can at least provide correct error message."""
    try:
        return logging.FileHandler(fileName)
    except OSError as e:
        logging.error("Failed to open logfile: [%s].", e)
        return None


def safeInfo(msg, *args, **kwargs):
    """Use tqdm redirect to not destroy progress bars"""
    rootlog = logging.getLogger("")
    try:
        from tqdm.contrib.logging import (  # pylint: disable=import-outside-toplevel
            logging_redirect_tqdm,
        )

        kwargs.setdefault("stacklevel", 2)
    except (ModuleNotFoundError, TypeError):
        rootlog.info(msg, *args, **kwargs)
        return

    with logging_redirect_tqdm():
        rootlog.info(msg, *args, **kwargs)


def configLogger(
    args: Namespace, fileLog: Optional[logging.FileHandler], counter: logCount
):
    """Setup logging"""
    syslog = False
    try:
        syslog = args.syslog is True
    except AttributeError:
        pass
    handler: List[Any]
    handler = [
        fileLog,
        counter,
    ]
    stderrh = logging.StreamHandler(stream=sys.stderr)
    if args.nocolor is False:
        formatter = colorlog.ColoredFormatter(
            logFormatColored,
            datefmt=logDateFormat,
            log_colors={
                "WARNING": "yellow",
                "ERROR": "red",
                "DEBUG": "cyan",
                "CRITICAL": "red",
            },
        )
        stderrh.setFormatter(formatter)
    if args.quiet is False:
        handler.append(stderrh)
    if syslog is True:
        handler.append(logging.handlers.SysLogHandler(address="/dev/log"))
    logging.basicConfig(
        level=setLogLevel(args.verbose),
        format=logFormat,
        datefmt=logDateFormat,
        handlers=handler,
    )


def hasFullBackup(args: Namespace) -> int:
    """Check if full backup file exists in target directory"""
    fullFiles = glob.glob(os.path.join(args.output, "*.full.data"))
    return len(fullFiles) > 0


def exists(args: Namespace, filePath: str) -> bool:
    """Check if file exists either remotely or locally."""
    if args.sshClient:
        return args.sshClient.exists(filePath)

    return os.path.exists(filePath)


def targetIsEmpty(args: Namespace) -> bool:
    """Check if target directory does not include an backup
    already (no .data or .data.partial files)"""
    if exists(args, args.output) and args.level in ("full", "copy", "auto"):
        dirList = glob.glob(os.path.join(args.output, "*.data*"))
        if len(dirList) > 0:
            return False

    return True


def getLatest(targetDir: str, search: str, key=None) -> List[str]:
    """get the last backed up file matching search
    from the backupset, used to find latest vm config,
    data files or data files by disk.
    """
    ret: List[str] = []
    try:
        files = glob.glob(os.path.join(targetDir, f"{search}"))
        files.sort(key=os.path.getmtime)

        if key is not None:
            ret.append(files[key])
        else:
            ret = files

        log.debug("Sorted data files: \n%s", pprint.pformat(ret))
        return ret
    except IndexError:
        return []


def hasQcowDisks(diskList: List[Any]) -> bool:
    """Check if the list of attached disks includes at least one
    qcow image based disk, else checkpoint handling can be
    skipped and backup module falls back to type copy"""
    for disk in diskList:
        if disk.format.startswith("qcow"):
            return True

    return False


def copy(args: Namespace, source: str, target: str) -> None:
    """Copy file, handle exceptions"""
    try:
        if args.sshClient:
            args.sshClient.copy(source, target)
        else:
            shutil.copyfile(source, target)
    except OSError as e:
        log.warning("Failed to copy [%s] to [%s]: [%s]", source, target, e)
    except sshError as e:
        log.warning("Remote copy from [%s] to [%s] failed: [%s]", source, target, e)


def remove(args: Namespace, file: str) -> None:
    """Remove file either locally or remote"""
    try:
        if args.sshClient:
            args.sshClient.run(f"rm -f {file}")
        else:
            os.remove(file)
        log.debug("Removed: [%s]", file)
    except FileNotFoundError:
        pass
    except OSError as e:
        log.warning("Failed to remove [%s]: [%s]", file, e)
    except sshError as e:
        log.warning("Remote remove failed: [%s]: [%s]", file, e)


def progressBar(total: int, desc: str, args: Namespace, count=0) -> tqdm:
    """Return tqdm object"""
    return tqdm(
        total=total,
        desc=desc,
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
        disable=args.noprogress,
        position=count,
        leave=False,
    )


def killProc(pid: int) -> bool:
    """Attempt kill PID"""
    log.debug("Killing PID: %s", pid)
    while True:
        try:
            os.kill(pid, signal.SIGTERM)
            return True
        except ProcessLookupError:
            return True


def getIdent(args: Namespace) -> Union[str, int]:
    """Used to get an unique identifier for target files,
    usually checkpoint name is used, but if no checkpoint
    is created, we use timestamp"""
    try:
        ident = args.cpt.name
    except AttributeError:
        ident = int(time())
    if args.level == "diff":
        ident = int(time())
    if args.level == "copy":
        ident = "copy"

    return ident


def dumpExtentJson(extents) -> str:
    """Dump extent object as json"""
    extList = []
    for extent in extents:
        ext = {}
        ext["start"] = extent.offset
        ext["length"] = extent.length
        ext["data"] = extent.data
        extList.append(ext)

    return json.dumps(extList, indent=4, sort_keys=True)


def dumpMetaData(dataFile: str, stream):
    """read metadata header"""
    with output.openfile(dataFile, "rb") as reader:
        _, _, length = stream.readFrame(reader)
        return stream.loadMetadata(reader.read(length))


def isCompressed(meta: Dict[str, str]) -> bool:
    """Return true if stream is compressed"""
    try:
        version = meta["stream-version"] == 2
    except KeyError:
        version = meta["streamVersion"] == 2

    if version:
        if meta["compressed"] is not False:
            return True

    return False



#!/usr/bin/python3
"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
from getpass import getuser
from argparse import _ArgumentGroup
from libvirtnbdbackup import __version__


def addRemoteArgs(opt: _ArgumentGroup) -> None:
    """Common remote backup arguments"""

    user = getuser() or None

    session = "qemu:///system"
    if user != "root":
        session = "qemu:///session"

    opt.add_argument(
        "-U",
        "--uri",
        default=session,
        required=False,
        type=str,
        help="Libvirt connection URI. (default: %(default)s)",
    )
    opt.add_argument(
        "--user",
        default=None,
        required=False,
        type=str,
        help="User to authenticate against libvirtd. (default: %(default)s)",
    )
    opt.add_argument(
        "--ssh-user",
        default=user,
        required=False,
        type=str,
        help=(
            "User to authenticate against remote sshd: "
            "used for remote copy of files. (default: %(default)s)"
        ),
    )
    opt.add_argument(
        "--ssh-port",
        default=22,
        required=False,
        type=int,
        help=(
            "Port to connect to remote sshd: "
            "used for remote copy of files. (default: %(default)s)"
        ),
    )
    opt.add_argument(
        "--password",
        default=None,
        required=False,
        type=str,
        help="Password to authenticate against libvirtd. (default: %(default)s)",
    )
    opt.add_argument(
        "-P",
        "--nbd-port",
        type=int,
        default=10809,
        required=False,
        help=(
            "Port used by remote NBD Service, should be unique for each"
            " started backup. (default: %(default)s)"
        ),
    )
    opt.add_argument(
        "-I",
        "--nbd-ip",
        type=str,
        default="",
        required=False,
        help=(
            "IP used to bind remote NBD service on"
            " (default: hostname returned by libvirtd)"
        ),
    )
    opt.add_argument(
        "--tls",
        action="store_true",
        required=False,
        help="Enable and use TLS for NBD connection. (default: %(default)s)",
    )
    opt.add_argument(
        "--tls-cert",
        type=str,
        default="/etc/pki/qemu/",
        required=False,
        help=(
            "Path to TLS certificates used during offline backup"
            " and restore. (default: %(default)s)"
        ),
    )


def addDebugArgs(opt: _ArgumentGroup) -> None:
    """Common debug arguments"""
    opt.add_argument(
        "-v",
        "--verbose",
        default=False,
        help="Enable debug output",
        action="store_true",
    )
    opt.add_argument(
        "-V",
        "--version",
        default=False,
        help="Show version and exit",
        action="version",
        version=__version__,
    )


def addLogArgs(opt, prog):
    """Logging related arguments"""
    try:
        HOME = os.environ["HOME"]
    except KeyError:
        HOME = "/tmp"
    opt.add_argument(
        "-L",
        "--logfile",
        default=f"{HOME}/{prog}.log",
        type=str,
        help="Path to Logfile (default: %(default)s)",
    )


def addLogColorArgs(opt):
    """Option to enable or disable colored output"""
    opt.add_argument(
        "--nocolor",
        default=False,
        help="Disable colored output (default: %(default)s)",
        action="store_true",
    )



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import glob
import logging
from argparse import Namespace

from libvirtnbdbackup import exceptions


log = logging.getLogger()


def _exists(args: Namespace) -> int:
    """Check for possible partial backup files"""
    partialFiles = glob.glob(f"{args.output}/*.partial")
    return len(partialFiles) > 0


def exists(args: Namespace) -> bool:
    """Check if target directory has an partial backup,
    makes backup utility exit errnous in case backup
    type is full or inc"""
    if args.level in ("inc", "diff") and args.stdout is False and _exists(args) is True:
        log.error("Partial backup found in target directory: [%s]", args.output)
        log.error("One of the last backups seems to have failed.")
        log.error("Consider re-executing full backup.")
        return True

    return False


def rename(targetFilePartial: str, targetFile: str) -> None:
    """After backup, move .partial file to real
    target file"""
    try:
        os.rename(targetFilePartial, targetFile)
    except OSError as e:
        raise exceptions.DiskBackupFailed(f"Failed to rename file: [{e}]") from e



#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from argparse import Namespace
from typing import Union
from libvirtnbdbackup import nbdcli
from libvirtnbdbackup import virt
from libvirtnbdbackup.virt.client import DomainDisk
from libvirtnbdbackup.qemu import util as qemu
from libvirtnbdbackup.objects import processInfo
from libvirtnbdbackup.nbdcli.exceptions import NbdClientException
from libvirtnbdbackup.exceptions import DiskBackupFailed


def setup(args: Namespace, disk: DomainDisk, remoteHost: str, port: int) -> processInfo:
    """Start background qemu-nbd process used during backup
    if domain is offline, in case of remote backup, initiate
    ssh session and start process on remote system."""
    bitMap: str = ""
    if args.level in ("inc", "diff"):
        bitMap = args.cpt.name
    socket = f"{args.socketfile}.{disk.target}"
    if remoteHost != "":
        logging.info(
            "Offline backup, starting remote NBD server, address: [%s:%s]",
            args.nbd_ip or remoteHost,
            port,
        )
        nbdProc = qemu.util(disk.target).startRemoteBackupNbdServer(
            args, disk, bitMap, port
        )
        logging.info("Remote NBD server started, PID: [%s].", nbdProc.pid)
        return nbdProc

    logging.info("Offline backup, starting local NBD server, socket: [%s]", socket)
    nbdProc = qemu.util(disk.target).startBackupNbdServer(
        disk.format, disk.path, socket, bitMap
    )
    logging.info("Local NBD Service started, PID: [%s]", nbdProc.pid)
    return nbdProc


def connect(  # pylint: disable=too-many-arguments
    args: Namespace,
    disk: DomainDisk,
    metaContext: str,
    remoteIP: str,
    port: int,
    virtClient: virt.client,
):
    """Connect to started nbd endpoint"""
    socket = args.socketfile
    if args.offline is True:
        socket = f"{args.socketfile}.{disk.target}"

    cType: Union[nbdcli.TCP, nbdcli.Unix]
    if virtClient.remoteHost != "":
        cType = nbdcli.TCP(disk.target, metaContext, remoteIP, args.tls, port)
    else:
        cType = nbdcli.Unix(disk.target, metaContext, socket)

    nbdClient = nbdcli.client(cType, args.no_sparse_detection)

    try:
        return nbdClient.connect()
    except NbdClientException as e:
        raise DiskBackupFailed(
            f"NBD endpoint: [{cType}]: connection failed: [{e}]"
        ) from e



#!/usr/bin/python3
"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import logging
import json
from typing import List, Dict, Tuple, IO
from libvirtnbdbackup import common as lib
from libvirtnbdbackup import output
from libvirtnbdbackup.output.exceptions import OutputException
from libvirtnbdbackup.exceptions import RestoreError
from libvirtnbdbackup.sparsestream.exceptions import StreamFormatException


def _parse(stream, sTypes, reader) -> Tuple[List, Dict]:
    """Read block offsets from backup stream image"""
    try:
        kind, start, length = stream.readFrame(reader)
        meta = stream.loadMetadata(reader.read(length))
    except StreamFormatException as errmsg:
        logging.error("Unable to read metadata header: %s", errmsg)
        raise RestoreError from errmsg

    if lib.isCompressed(meta):
        logging.error("Mapping compressed images currently not supported.")
        raise RestoreError

    assert reader.read(len(sTypes.TERM)) == sTypes.TERM

    dataRanges: List = []
    count: int = 0
    while True:
        kind, start, length = stream.readFrame(reader)
        if kind == sTypes.STOP:
            dataRanges[-1]["nextBlockOffset"] = None
            break

        blockInfo = {}
        blockInfo["count"] = count
        blockInfo["offset"] = reader.tell()
        blockInfo["originalOffset"] = start
        blockInfo["nextOriginalOffset"] = start + length
        blockInfo["length"] = length
        blockInfo["data"] = kind == sTypes.DATA
        blockInfo["file"] = os.path.abspath(reader.name)
        blockInfo["inc"] = meta["incremental"]

        if kind == sTypes.DATA:
            reader.seek(length, os.SEEK_CUR)
            assert reader.read(len(sTypes.TERM)) == sTypes.TERM

        nextBlockOffset = reader.tell() + sTypes.FRAME_LEN
        blockInfo["nextBlockOffset"] = nextBlockOffset
        dataRanges.append(blockInfo)
        count += 1

    return dataRanges, meta


def get(args, stream, sTypes, dataFiles: List) -> List:
    """Get data ranges for each file specified"""
    dataRanges = []
    for dFile in dataFiles:
        try:
            reader = output.openfile(dFile, "rb")
        except OutputException as e:
            logging.error("[%s]: [%s]", dFile, e)
            raise RestoreError from e

        Range, meta = _parse(stream, sTypes, reader)
        if Range is False or meta is False:
            logging.error("Unable to read meta header from backup file.")
            raise RestoreError("Invalid header")
        dataRanges.extend(Range)

        if args.verbose is True:
            logging.info(json.dumps(dataRanges, indent=4))
        else:
            logging.info(
                "Parsed [%s] block offsets from file [%s]", len(dataRanges), dFile
            )
        reader.close()

    return dataRanges


def dump(tfile: IO, dataRanges: List) -> bool:
    """Dump block map to temporary file"""
    try:
        tfile.write(json.dumps(dataRanges, indent=4).encode())
        return True
    except OSError as e:
        logging.error("Unable to write blockmap file: %s", e)
        return False
